#include <stdio.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/stat.h>

#define PATH_BOOTMAP  "/boot/bootmap"
#define PATH_BOOTCONF "/boot/boot.conf"

int verbose=0;
int ofd=0;

int inline map(int fd, int ptr)
{
  int p;
  
  p=ptr;
  
  if(ioctl(fd,1,&p)==-1)
  {
    perror("ioctl FIBMAP");
    close(fd);
    exit(1);
  }

  return p;
}

int create_rootentry(char *dev,char *mapfile)
{
  int blocks[64];
  int ptr, fd, blk;

  if(verbose)
    fprintf(stderr,"Creating root map...\n");
  
  fd = open(mapfile,O_RDONLY,0);
  if(fd==-1)
  {
    perror("bootmap: rootentry: file open");
    return 1;
  }
    
  memset(&blocks[0], 0, sizeof(blocks));

  memcpy(&blocks[0], "MKLD", 4);
  blocks[1]=0xCCAA5533;
  blocks[2]=1024;			/* Size of blocks in bytes */
  ptr=0;
  
  do
  {
    blk=blocks[ptr+3]=map(fd,ptr);
    ptr++;
  }
  while(blk);

  close(fd);

  if(verbose)
    fprintf(stderr,"Writing to device...\n");
    
  fd = open(dev, O_WRONLY, 0);
  if(fd == -1)
  {
    perror("bootmap: rootentry: dev open");
    return 1;
  }
    
  write(fd, &blocks[0], sizeof(blocks));
  close(fd);
  return 0;
}

int bitmap(int *buffer,char *path)
{
  int fd, ptr, n, i, oldblk, start, blk;
  int totlen = 0;
  
  fd = open(path,O_RDONLY,0);
  if(fd == -1)
  {
    perror("bootmap: bitmap: open");
    return 1;
  }

  n = 0;
  oldblk = 0;
  ptr = 0;

  do
  {
    if(oldblk == 0)
      start = oldblk = map(fd, ptr++);
    blk = map(fd, ptr++);
    if(blk != oldblk+1)
    {
      buffer[n++]=start;
      buffer[n++]=oldblk-start+1;
      totlen += oldblk - start+1;
      start = blk;
    }
    oldblk = blk;
    if(n>=53)
    {
      close(fd);
      printf("Kernel image %s is in too many bits - ignoring\n",path);
      return 1;
    }
  }
  while(blk);

  totlen *= 1024;
  if(verbose)
    printf("%s: total length is %d bytes\n",path,totlen);
  
  close(fd);

  return 0;
}

int add_one_dev(char *name, char *path, char *pw, int rootdev)
{
  int nopath, nopw, r=0;
  char buffer[256];
  char b[8];

  if(!ofd)
  {
    ofd = open(PATH_BOOTMAP,O_CREAT|O_WRONLY);
    if(ofd == -1)
    {
      ofd = 0;
      perror("bootmap: add_one_dev: open");
      return 1;
    }
  }

  nopath = strcmp(path,"none");
  pw[strlen(pw)-1]=0;
  nopw   = strcmp(pw,"none");

  memset(buffer, 0, sizeof(buffer));
  memset(b, 0, 8);
  strcpy(b, pw);
  
  strcpy(buffer, name);

  if(nopath)
  {
    *(int *)(buffer+40) = rootdev;
    r = bitmap((int *)(buffer+44), path);
  }

  if(r)
    return r;

  if(nopw)
  {
    buffer[32] = buffer[4] ^ 'w' ^ b[0];
    buffer[33] = buffer[5] ^ 'o' ^ b[1];
    buffer[34] = buffer[6] ^ 'r' ^ b[2];
    buffer[35] = buffer[7] ^ 'd' ^ b[3];
    buffer[36] = buffer[0] ^ 'P' ^ b[4];
    buffer[37] = buffer[1] ^ 'a' ^ b[5];
    buffer[38] = buffer[2] ^ 's' ^ b[6];
    buffer[39] = buffer[3] ^ 's' ^ b[7];
  }

  if(write(ofd,buffer,sizeof(buffer))!=sizeof(buffer))
  {
    perror("bootmap: add_one_dev: write");
    return 1;
  }
  
  return 0;
}

int create_entries(int rootdev)
{
  FILE *f;
  char line[256];
  
  f=fopen(PATH_BOOTCONF,"r");

  do
  {
    char *name,*path,*pw;
    fgets(line, 256, f);
    if(feof(f) || line[0]=='#' || (strrchr(line,' ')==NULL && strrchr(line,'\t')==NULL))
      continue;

    name=strtok(line," \t");
    if(!name) continue;
    path=strtok(NULL," \t");
    if(!path) continue;
    pw  =strtok(NULL," \t");
    if(!pw)   continue;
    
    if(add_one_dev(name, path, pw, rootdev))
      break;
  }
  while(!feof(f));

  if(ofd)
    close(ofd);
    
  fclose(f);
}

int main(int argc,char *argv[])
{
  int rootdev;
  struct stat st;
  
  if(argc<2)
  {
    fprintf(stderr,"Usage: bootmap <device-file>\n");
    exit(1);
  }

  if(stat("/", &st)== -1)
  {
    perror("stat: /");
    exit(1);
  }
  
  rootdev = st.st_dev;
printf("root dev is %X\n",rootdev);
  
  create_entries(rootdev);

  create_rootentry(argv[1], PATH_BOOTMAP);

  chown(PATH_BOOTMAP, 0, 0);
  chmod(PATH_BOOTMAP, 0400);

  chown(PATH_BOOTCONF, 0, 0);
  chmod(PATH_BOOTCONF, 0600);
  
  exit(0);
}
